%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "set.h"
#include "nodes.h"
#include "config.h"

#define MAX_LINE_LEN	2048

#define YYDEBUG 1

static struct nodelist *nl;
char *input_buf_start, *input_buf_end;

static struct nodelist *nodelist_create(const size_t maxsize);
static int nodelist_insert(struct nodelist *nl, struct nodegroup *ng);
static struct nodelist *nodelist_union(struct nodelist *a, struct nodelist *b);

static int cmp_int(const void *a, const void *b);

int yyerror(char *s);

extern int yylex();
%}

%union {
	int val;
	char *fmt;
	struct nodegroup *group;
	struct nodelist *list;
	struct set *iset;
}

%token <val> ERROR
%token <val> INTEGER
%token <fmt>  NODE FORMAT
%type <list> nodelist
%type <iset> intset
%type <group> nodegroup

%%

nodelist:
	ERROR						{
									YYABORT;
								}
	| nodegroup					{ 
									nl = nodelist_create(MAX_NODES);

									if (nodelist_insert(nl, $1) < 0){
										YYABORT;
									}
									free($1);

									$$ = nl;
								}
	| nodegroup nodelist 		{ 
									if (nodelist_insert($2, $1) < 0){
										YYABORT;
									}
									free($1);

									$$ = $2;
								}
	;

nodegroup:
	NODE						{ 
									struct nodegroup *ng;
	
									ng = malloc(sizeof(struct nodegroup));
									if (ng == NULL){
										YYABORT;
									}
									ng->fmt = $1;
									if (ng->fmt == NULL){
										free(ng);
										YYABORT;
									}
									ng->nodes = NULL;

									$$ = ng;
								}
	| FORMAT '@' intset			{
									struct nodegroup *ng;
				
									ng = malloc(sizeof(struct nodegroup));
									if (ng == NULL){
										YYABORT;
									}
									ng->fmt = $1;
									if (ng->fmt == NULL){
										free(ng);
										YYABORT;
									}
									ng->nodes = $3;

									$$ = ng;
								}
	;

intset:
	INTEGER						{	
									$$ = set_create(MAX_NODES, 
												sizeof(int), 
											cmp_int);
									if ($$ == NULL){
										YYABORT;
									}

									if (set_insert($$, &$1) < 0){
										YYABORT;
									}
								}
	| INTEGER '-' INTEGER		{
									int i;
			
									if ($1 > $3){
										yyerror("Misordered range");
										YYABORT;
									}

									$$ = set_create(MAX_NODES, sizeof(int), 
										            cmp_int);

									if ($$ == NULL){
										YYABORT;
									}

									for (i = $1; i <= $3; i++){
										if (set_insert($$, &i) < 0){
											YYABORT;
										}
									}
								}
	| INTEGER ',' intset		{	
									if (set_insert($3, &$1) < 0){
										YYABORT;
									}
									
									$$ = $3;
								}
	| INTEGER'-'INTEGER','intset{	
									int i;
				
									if ($1 > $3){
										yyerror("Misordered range");
										YYABORT;
									}

									for (i = $1; i <= $3; i++){
										if (set_insert($5, &i) < 0){
											YYABORT;
										}
									}
						
									$$ = $5;
								}
	;

%%

int yyerror(char *s)
{
	fprintf(stderr, "%s\n", s);
	return 0;
}

/*@ 
   nodelist_create - create a new nodelist

Input arguments:
+ maxsize - The maximum size of the nodelist

Output arguments:
None

Returns:
The new nodelist
@*/
static struct nodelist *nodelist_create(const size_t maxsize)
{
	struct nodelist *nl;

	nl = malloc(sizeof(struct nodelist));
	if (nl == NULL)
		return NULL;
	
	nl->list = malloc(sizeof(struct nodegroup)*maxsize);
	if (nl->list == NULL){
		free(nl);
		return NULL;
	}

	nl->reduced = 1;
	nl->size = 0;
	nl->maxsize = maxsize;

	return nl;

}

/*@ 
   nodelist_destroy - destroy a nodelist 

Input arguments:
+ nl - The nodelist to be destroyed

Output arguments:
None
@*/
void nodelist_destroy(struct nodelist *nl)
{
	int i;

	if (nl == NULL)
		return;

	for (i = 0; i < nl->size; i++){
		set_destroy(nl->list[i].nodes);
	}
	free(nl->list);

	free(nl);
}

/* an integer comparison function with qsort semantics */
static int cmp_int(const void *a, const void *b)
{
	int A = *(int *)a, B = *(int *)b;

	if (A < B)
		return -1;
	if (A > B)
		return 1;
    
	return 0;
}

/* an nodegroup comparison function with qsort semantics */
int cmp_nodegroup(const void *a, const void *b)
{
	const struct nodegroup *m = a;
	const struct nodegroup *n = b;

	return strcmp(m->fmt, n->fmt);
}

/*@ 
   nodelist_reduce - reduce a nodelist by grouping common elements

Input arguments:
+ nl - the nodelist to be reduced

Output arguments:
None

Returns:
0 on success, -1 on failure
@*/
int nodelist_reduce(struct nodelist *nl)
{
	int i;
	char buf[1024];

	qsort(nl->list, nl->size, sizeof(struct nodegroup), cmp_nodegroup);

	strncpy(buf, nl->list[0].fmt, 1024);
	for (i = 1; i < nl->size; i++){
		if (strcmp(nl->list[i].fmt, buf) == 0){
			struct set *tmp;

			if (nl->list[i].nodes != NULL && nl->list[i-1].nodes != NULL){
				tmp = set_union(nl->list[i].nodes, nl->list[i-1].nodes);
				if (tmp == NULL)
					return -1;

				set_destroy(nl->list[i].nodes);
				set_destroy(nl->list[i-1].nodes);
				nl->list[i-1].nodes = tmp;
			} else if (nl->list[i].nodes == NULL)
				nl->list[i-1].nodes = nl->list[i].nodes;

			memmove(nl->list+i, nl->list+i+1, 
				sizeof(struct nodegroup)*(nl->size-i-1));
			nl->size--, i--;
		} else {
			strncpy(buf, nl->list[i].fmt, 1024);
		}
	}

	nl->reduced = 1;

	return 0;
}

/*@ 
   nodelist_reset - reset a nodelist for iteration

Input arguments:
+ nl - the nodelist to be reset

Output arguments:
None
@*/
void nodelist_reset(struct nodelist *nl)
{
	int i;

	for (i = 0; i < nl->size; i++){
		set_reset(nl->list[i].nodes);
	}

	nl->idx = 0;
}

/*@ 
   nodelist_next - retrieve the next node in the list

Input arguments:
+ nl - the nodelist from which to retrieve the next node
- bufsize - the size of buf

Output arguments:
+ buf - the buffer into which the next node name is stored

Returns:
buf on success, NULL on failure
@*/
char *nodelist_next(struct nodelist *nl, char *buf, size_t bufsize)
{
	char *fmt;
	int *node = NULL;

	if ((!nl->reduced) && (nodelist_reduce(nl) < 0))
		return NULL;

	do {
		if (nl->idx >= nl->size){
			return NULL;
		}

		fmt = nl->list[nl->idx].fmt;
		/* This is an absolute node name */
		if (nl->list[nl->idx].nodes == NULL){
			strncpy(buf, nl->list[nl->idx].fmt, bufsize);
			nl->idx++;
			return buf;
		} else {
			node = set_next(nl->list[nl->idx].nodes);
			if (node == NULL)
				nl->idx++;
		}
		
	} while (node == NULL);

	snprintf(buf, bufsize, fmt, *node);

	return buf;
}

/*@ 
   nodelist_next - retrieve the next compact node specifier in the list

Input arguments:
+ nl - the nodelist from which to retrieve the next node
- bufsize - the size of buf

Output arguments:
+ buf - the buffer into which the next node specifier is stored

Returns:
buf on success, NULL on failure or if the output was truncated 
@*/
char *nodelist_next_cmpt(struct nodelist *nl, char *buf, size_t bufsize)
{
	char *fmt, *ptr = buf, *limit = buf+bufsize;
	int *node = NULL;
	int start, end;
	int printed = 0; /* Indicates if _any_ printing has been done so that */ 
                     /* commas only go in between ranges and numbers */
	int len, fmtlen;
	struct nodegroup *ng;

	if (nl->idx >= nl->size){
		return NULL;
	}

	if ((!nl->reduced) && (nodelist_reduce(nl) < 0))
		return NULL;

	ng = &(nl->list[nl->idx]);
	fmt = ng->fmt;
	fmtlen = strlen(fmt);
	/* This is an absolute node name */
	if (ng->nodes == NULL){
		if (strlen(fmt) >= bufsize){
			nl->idx++;
			return NULL;
		}
		strcpy(ptr, fmt);
		nl->idx++;
		return buf;
	}
		
	ptr[0] = '\0';

	if (strlen(fmt)+1 >= bufsize){
		nl->idx++;
		return NULL;
	}
	strcpy(ptr, fmt);
	ptr[fmtlen] = '@';
	ptr += fmtlen+1;
	
	set_reset(ng->nodes);
	/* ng->nodes must be non-empty */
	start = end = *(int *)set_next(ng->nodes);
	for (node = set_next(ng->nodes); node != NULL; node = set_next(ng->nodes)){
		if (*node == end+1){
			end++;
		} else {
			if (printed){
				len = snprintf(ptr, limit-ptr, ",");
				
				/* The string was truncated */
				if (len < 0 || len > limit-ptr){
					nl->idx++;
					return NULL;
				}

				ptr += len;
			}

			if (end > start){
				len = snprintf(ptr, limit-ptr, "%d-%d", start, end);

				if (len < 0 || len > limit-ptr){
					nl->idx++;
					return NULL;
				}

				ptr += len;

				start = end = *node;

				printed = 1;
			} else {
				len = snprintf(ptr, limit-ptr, "%d", end);

				if (len < 0 || len > limit-ptr){
					nl->idx++;
					return NULL;
				}

				ptr += len;

				start = end = *node;

				printed = 1;
			}
		}
	}

	if (printed){
		len = snprintf(ptr, limit-ptr, ",");
				
		if (len < 0 || len > limit-ptr){
			nl->idx++;
			return NULL;
		}

		ptr += len;
	}

	if (end > start){
		len = snprintf(ptr, limit-ptr, "%d-%d", start, end);

		if (len < 0 || len > limit-ptr){
			nl->idx++;
			return NULL;
		}
	} else {
		len = snprintf(ptr, limit-ptr, "%d", end);

		if (len < 0 || len > limit-ptr){
			nl->idx++;
			return NULL;
		}
	}
	nl->idx++;

	return buf;
}

/*@ 
   nodelist_parse - parse a nodelist from a buffer

Input arguments:
+ list - The buffer to parse

Output arguments:
None

Returns:
A new nodelist of the parsed buffer if successful, NULL otherwise

Notes:
The new nodelist must be freed by the caller with nodelist_destroy
@*/
struct nodelist *nodelist_parse(char *list)
{
	input_buf_start = list;
	input_buf_end = list+strlen(list);
	
	if (getenv("DEBUG"))
		yydebug = 1;

	if (yyparse() != 0){
		nodelist_destroy(nl);
		return NULL;
	}

	if (nodelist_reduce(nl) < 0){
		nodelist_destroy(nl);
		return NULL;
	}

	return nl;
}

/*@ 
   nodelist_parse_file - parse a nodelist from a file

Input arguments:
+ fn - the filename of the file to parse

Output arguments:
None

Returns:
A new nodelist of the parsed buffer if successful, NULL otherwise

Notes:
The new nodelist must be freed by the caller with nodelist_destroy
@*/
struct nodelist *nodelist_parse_file(char *fn)
{
	FILE *f;
	int reduced = 0;
	struct nodelist *nl = NULL, *tmp, *curr;
	char line[MAX_LINE_LEN];

	f = fopen(fn, "r");
	if (f == NULL)
		return NULL;

	while (fgets(line, MAX_LINE_LEN, f) != NULL){
		curr = nodelist_parse(line);
		reduced = 0;

		if (curr == NULL){
			nodelist_destroy(nl);
			return NULL;
		}

		if (nl == NULL){
			nl = curr;
		} else {
			tmp = nl;
			nl = nodelist_union(tmp, curr);
			nodelist_destroy(tmp);
			nodelist_destroy(curr);
			if (nl == NULL)
				return NULL;
			reduced = 1;
		}
	}

	if (!reduced && nodelist_reduce(nl) < 0){
		nodelist_destroy(nl);
		return NULL;
	}
		
	return nl;
}

/*@ 
   nodelist_insert - add a nodegroup to a nodelist

Input arguments:
+ nl - the nodelist
- ng - the nodegroup

Output arguments:
None

Returns:
0 if successful, -1 otherwise
@*/
static int nodelist_insert(struct nodelist *nl, struct nodegroup *ng)
{
	if (nl == NULL || ng == NULL)
		return -1;

	if (nl->size >= nl->maxsize){
		nl->size = nl->maxsize;
		return -1;
	}

	memcpy(&(nl->list[nl->size]), ng, sizeof(struct nodegroup));

	nl->size++;
	nl->reduced = 0;

	return 0;
}

/*@ 
   nodelist_insert_node - add single node to a nodelist

Input arguments:
+ nl - the nodelist
- nodename - the node name

Output arguments:
None

Returns:
0 if successful, -1 otherwise
@*/
int nodelist_insert_node(struct nodelist *nl, char *nodename)
{
	struct nodegroup *ng;
	
	if (nl == NULL || nodename == NULL)
		return -1;

	if (nl->size >= nl->maxsize){
		nl->size = nl->maxsize;
		return -1;
	}

	ng = &(nl->list[nl->size]);
	ng->fmt = nodename;
	ng->nodes = NULL;

	nl->size++;
	nl->reduced = 0;

	return 0;
}

/*@ 
   nodelist_union - join two nodelists

Input arguments:
+ a - the first nodelist
- b - the second nodelist

Output arguments:
None

Returns:
A new nodelist that is the union of a and b if successful, NULL otherwise
@*/
static struct nodelist *nodelist_union(struct nodelist *a, struct nodelist *b)
{
	struct nodelist *c;

	if (a == NULL || b == NULL)
		return NULL;

	c = nodelist_create(a->maxsize+b->maxsize);
	if (c == NULL)
		return NULL;

	memcpy(c->list, a->list, a->size*(sizeof(struct nodegroup)));
	memcpy(c->list+a->size, b->list, b->size*(sizeof(struct nodegroup)));
	c->size = a->size + b->size;

	if (nodelist_reduce(c) < 0){
		nodelist_destroy(c);
		return NULL;
	}

	return c;
}

/*@
  nodelist_exists - tests if a node is in a nodelist

Input arguments:
+ nl - the nodelist
- nodename - the node

Output arguments:
None

Returns:
1 if the node is in the nodelist, 0 otherwise

Notes:
The nodelist must be reset after this function.

@*/
extern int nodelist_exists(struct nodelist *nl, char *nodename)
{
	char buf[1024];

	if ((!nl->reduced) && (nodelist_reduce(nl) < 0))
		return 0;

	nodelist_reset(nl);
	
	while (nodelist_next(nl, buf, 1024) != NULL){
		if (strcmp(nodename, buf) == 0)
			return 1;
	}

	return 0;
}

