/* A wrapper that runs mpirun with the right arguments */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stropts.h>
#include <termio.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include "ptools.h"
#include "debug.h"
#include "nodes.h"

#define BUF_SIZE	1024

/*N machine_args
+	-m <machine_file> - 
		Specify a file with a list of destination machines.  It is incorrect
		to use this with the -M option.  This option must preceed all others.

.	-M <machine_list> -
		Specify a list of destination machines.  The entire list must be
		surrounded by quotes.  It is incorrect to use this with the -m option.
		This option must preceed all others.
.   -all -
        Specify all machines in the mpd ring.
 N*/

/*N machine_envs
-	MACHINE_FILE -
		If MACHINE_FILE is set, the command will attempt to open the file 
		whose name is stored in MACHINE_FILE.  This will be used if neither 
		the -m or -M options are given.
  N*/

char *screen_width()
{ 
    struct winsize ws;
    static char buf[24];
    char *tmp = getenv("COLUMNS");

    if (tmp != NULL)
        return tmp;

    if (ioctl(STDOUT_FILENO, TIOCGWINSZ, &ws) != 0)
        return "80";

    snprintf(buf, sizeof(buf), "%d", ws.ws_col);

    return buf;
}


int main(int argc, char *argv[])
{
	int num_machines;
	int i, j, status, pid;
    int all = 0, need_cols = 0;
    int optind;
	char buf[BUF_SIZE];
	char cmd[BUF_SIZE];
	char **mpirun_argv;
    char *tmp, *tmpfn;

    if (nodes_init(argc, argv, 1, &tmpfn, &num_machines, &all, &optind) < 0)
        return 1;

	/*****************************************/
	/* Get the arguments ready to run mpirun */
	/*****************************************/
	mpirun_argv = malloc((argc+8)*sizeof(char *));
	if (mpirun_argv == NULL){
		perror("Unable to allocate new argument vector");
		return 1;
	}

	if (strlen(argv[0]) + sizeof(".mpi") >= BUF_SIZE){
		fprintf(stderr, "Unknown command '%s'\n", argv[0]);
		return 1;
	}
	strcpy(cmd, argv[0]);
    tmp = strrchr(cmd, '/');
    if (tmp == NULL) {
        if (strcmp(cmd, "ptfps") == 0)
            need_cols = 1;
    } else {
        if (strcmp(tmp + 1, "ptfps") == 0)
            need_cols = 1;
    }

	strcat(cmd, ".mpi");

    j = 0;
    mpirun_argv[j++] = "mpirun";
	mpirun_argv[j++] = "-np";
	snprintf(buf, BUF_SIZE, "%d", num_machines);
	mpirun_argv[j++] = buf;

    if (!all) {
        mpirun_argv[j++] = "-m";
        mpirun_argv[j++] = tmpfn;
    }
    mpirun_argv[j++] = cmd;
    if (need_cols)
        mpirun_argv[j++] = screen_width();
    for (i = optind; i <= argc; i++)
        mpirun_argv[j+(i-optind)] = argv[i];
    
	/**************/
	/* Run mpirun */
	/**************/
	pid = fork();
	if (pid < 0){
		perror("Unable to fork");
		return 1;
	}
	if (pid == 0){
		pt_debug("Executing '%s'\n", cmd);
		execvp(mpirun_argv[0], mpirun_argv);
		return 1;
	}
	wait(&status);
	if (!WIFEXITED(status))
		fprintf(stderr, "%s exited abnormally\n", mpirun_argv[0]);

	pt_debug("%s exited with code %d\n", cmd, WEXITSTATUS(status));

	/****************************************/
	/* Remove the intermediate machine file */
	/****************************************/
    unlink(tmpfn);

	return 0;
}
